"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var path = require("path");
var fs = require("fs");
var defaultOptions = {
    target: ts.ScriptTarget.Latest,
    module: ts.ModuleKind.CommonJS,
    jsx: ts.JsxEmit.React,
};
/**
 * Parses a file with default TS options
 * @param filePath component file that should be parsed
 */
function parse(filePath) {
    return withCompilerOptions(defaultOptions).parse(filePath);
}
exports.parse = parse;
/**
 * Constructs a parser for a default configuration.
 */
function withDefaultConfig() {
    return withCompilerOptions(defaultOptions);
}
exports.withDefaultConfig = withDefaultConfig;
/**
 * Constructs a parser for a specified tsconfig file.
 */
function withCustomConfig(tsconfigPath) {
    var configJson = JSON.parse(fs.readFileSync(tsconfigPath, 'utf8'));
    var basePath = path.dirname(tsconfigPath);
    var _a = ts.convertCompilerOptionsFromJson(configJson.compilerOptions, basePath, tsconfigPath), options = _a.options, errors = _a.errors;
    if (errors && errors.length) {
        throw errors[0];
    }
    return withCompilerOptions(options);
}
exports.withCustomConfig = withCustomConfig;
/**
 * Constructs a parser for a specified set of TS compiler options.
 */
function withCompilerOptions(compilerOptions) {
    return {
        parse: function (filePath) {
            var program = ts.createProgram([filePath], compilerOptions);
            var parser = new Parser(program);
            var checker = program.getTypeChecker();
            var sourceFile = program.getSourceFile(filePath);
            var moduleSymbol = checker.getSymbolAtLocation(sourceFile);
            var exports = checker.getExportsOfModule(moduleSymbol);
            var components = exports
                .map(function (exp) { return parser.getComponentInfo(exp, sourceFile); })
                .filter(function (comp) { return comp; });
            return components;
        }
    };
}
exports.withCompilerOptions = withCompilerOptions;
var defaultJSDoc = {
    fullComment: '',
    tags: {},
    description: ''
};
var Parser = (function () {
    function Parser(program) {
        this.checker = program.getTypeChecker();
    }
    Parser.prototype.getComponentInfo = function (exp, source) {
        var type = this.checker.getTypeOfSymbolAtLocation(exp, exp.valueDeclaration);
        var propsType = this.extractPropsFromTypeIfStatelessComponent(type);
        if (!propsType) {
            propsType = this.extractPropsFromTypeIfStatefulComponent(type);
        }
        if (propsType) {
            var componentName = computeComponentName(exp, source);
            var defaultProps = this.extractDefaultPropsFromComponent(exp, source);
            var props = this.getPropsInfo(propsType, defaultProps);
            return {
                displayName: componentName,
                description: this.findDocComment(exp).fullComment,
                props: props
            };
        }
        return null;
    };
    Parser.prototype.extractPropsFromTypeIfStatelessComponent = function (type) {
        var callSignatures = type.getCallSignatures();
        if (callSignatures.length) {
            // Could be a stateless component.  Is a function, so the props object we're interested
            // in is the (only) parameter.
            for (var _i = 0, callSignatures_1 = callSignatures; _i < callSignatures_1.length; _i++) {
                var sig = callSignatures_1[_i];
                var params = sig.getParameters();
                if (params.length === 0) {
                    continue;
                }
                // Maybe we could check return type instead,
                // but not sure if Element, ReactElement<T> are all possible values
                var propsParam = params[0];
                if (propsParam.name === 'props' || params.length === 1) {
                    return propsParam;
                }
            }
        }
        return null;
    };
    Parser.prototype.extractPropsFromTypeIfStatefulComponent = function (type) {
        var constructSignatures = type.getConstructSignatures();
        if (constructSignatures.length) {
            // React.Component. Is a class, so the props object we're interested
            // in is the type of 'props' property of the object constructed by the class.
            for (var _i = 0, constructSignatures_1 = constructSignatures; _i < constructSignatures_1.length; _i++) {
                var sig = constructSignatures_1[_i];
                var instanceType = sig.getReturnType();
                var props = instanceType.getProperty('props');
                if (props) {
                    return props;
                }
            }
        }
        return null;
    };
    Parser.prototype.getPropsInfo = function (propsObj, defaultProps) {
        var _this = this;
        if (defaultProps === void 0) { defaultProps = {}; }
        var propsType = this.checker.getTypeOfSymbolAtLocation(propsObj, propsObj.valueDeclaration);
        var propertiesOfProps = propsType.getProperties();
        var result = {};
        propertiesOfProps.forEach(function (prop) {
            var propName = prop.getName();
            // Find type of prop by looking in context of the props object itself.
            var propType = _this.checker.getTypeOfSymbolAtLocation(prop, propsObj.valueDeclaration);
            var propTypeString = _this.checker.typeToString(propType);
            var isOptional = (prop.getFlags() & ts.SymbolFlags.Optional) !== 0;
            var jsDocComment = _this.findDocComment(prop);
            var defaultValue = null;
            if (defaultProps[propName] !== undefined) {
                defaultValue = { value: defaultProps[propName] };
            }
            else if (jsDocComment.tags.default) {
                defaultValue = { value: jsDocComment.tags.default };
            }
            result[propName] = {
                required: !isOptional,
                type: { name: propTypeString },
                description: jsDocComment.fullComment,
                defaultValue: defaultValue
            };
        });
        return result;
    };
    Parser.prototype.findDocComment = function (symbol) {
        var _this = this;
        var comment = this.getFullJsDocComment(symbol);
        if (comment.fullComment) {
            return comment;
        }
        var rootSymbols = this.checker.getRootSymbols(symbol);
        var commentsOnRootSymbols = rootSymbols
            .filter(function (x) { return x !== symbol; })
            .map(function (x) { return _this.getFullJsDocComment(x); })
            .filter(function (x) { return !!x.fullComment; });
        if (commentsOnRootSymbols.length) {
            return commentsOnRootSymbols[0];
        }
        return defaultJSDoc;
    };
    /**
     * Extracts a full JsDoc comment from a symbol, even
     * though TypeScript has broken down the JsDoc comment into plain
     * text and JsDoc tags.
     */
    Parser.prototype.getFullJsDocComment = function (symbol) {
        // in some cases this can be undefined (Pick<Type, 'prop1'|'prop2'>)
        if (symbol.getDocumentationComment === undefined) {
            return defaultJSDoc;
        }
        var mainComment = ts.displayPartsToString(symbol.getDocumentationComment());
        var tags = symbol.getJsDocTags() || [];
        var tagComments = [];
        var tagMap = {};
        tags.forEach(function (tag) {
            var trimmedText = (tag.text || '').trim();
            var currentValue = tagMap[tag.name];
            tagMap[tag.name] = currentValue ? currentValue + '\n' + trimmedText : trimmedText;
            if (tag.name !== 'default') {
                tagComments.push(formatTag(tag));
            }
        });
        return ({
            fullComment: (mainComment + '\n' + tagComments.join('\n')).trim(),
            tags: tagMap,
            description: mainComment
        });
    };
    Parser.prototype.extractDefaultPropsFromComponent = function (symbol, source) {
        var _this = this;
        var possibleStatements = source.statements.filter(function (statement) { return _this.checker.getSymbolAtLocation(statement.name) === symbol; });
        if (!possibleStatements.length) {
            return {};
        }
        var statement = possibleStatements[0];
        if (statementIsClassDeclaration(statement) && statement.members.length) {
            var possibleDefaultProps = statement.members.filter(function (member) { return member.name && getPropertyName(member.name) === 'defaultProps'; });
            if (!possibleDefaultProps.length) {
                return {};
            }
            var defaultProps = possibleDefaultProps[0];
            var initializer = defaultProps.initializer;
            var properties = initializer.properties;
            var propMap = properties.reduce(function (acc, property) {
                var literalValue = getLiteralValueFromPropertyAssignment(property);
                if (typeof literalValue === 'string') {
                    acc[getPropertyName(property.name)] = getLiteralValueFromPropertyAssignment(property);
                }
                return acc;
            }, {});
            return propMap;
        }
        return {};
    };
    return Parser;
}());
function statementIsClassDeclaration(statement) {
    return !!statement.members;
}
function getPropertyName(name) {
    switch (name.kind) {
        case ts.SyntaxKind.NumericLiteral:
        case ts.SyntaxKind.StringLiteral:
        case ts.SyntaxKind.Identifier:
            return name.text;
        case ts.SyntaxKind.ComputedPropertyName:
            return name.getText();
        default:
            return null;
    }
}
function getLiteralValueFromPropertyAssignment(property) {
    var initializer = property.initializer;
    switch (initializer.kind) {
        case ts.SyntaxKind.FalseKeyword:
            return 'false';
        case ts.SyntaxKind.TrueKeyword:
            return 'true';
        case ts.SyntaxKind.StringLiteral:
            return initializer.text.trim();
        case ts.SyntaxKind.NumericLiteral:
            return "" + initializer.text;
        case ts.SyntaxKind.NullKeyword:
            return 'null';
        case ts.SyntaxKind.Identifier:
            // can potentially find other identifiers in the source and map those in the future
            return initializer.text === 'undefined' ? 'undefined' : null;
        case ts.SyntaxKind.ObjectLiteralExpression:
            // return the source text for an object literal
            return initializer.getText();
        default:
            return null;
    }
}
function formatTag(tag) {
    var result = '@' + tag.name;
    if (tag.text) {
        result += ' ' + tag.text;
    }
    return result;
}
function computeComponentName(exp, source) {
    var exportName = exp.getName();
    if (exportName === 'default') {
        // Default export for a file: named after file
        return path.basename(source.fileName, path.extname(source.fileName));
    }
    else {
        return exportName;
    }
}
//# sourceMappingURL=parser.js.map